﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Threading;
using HeuristicLab.Algorithms.MemPR.Interfaces;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.MemPR.Permutation.LocalSearch {
  [Item("Exhaustive Hillclimber (permutation)", "", ExcludeGenericTypeInfo = true)]
  [StorableClass]
  public class ExhaustiveHillClimb<TContext> : NamedItem, ILocalSearch<TContext>
      where TContext : ISingleSolutionHeuristicAlgorithmContext<ISingleObjectiveHeuristicOptimizationProblem, Encodings.PermutationEncoding.Permutation>,
                       IEvaluationServiceContext<Encodings.PermutationEncoding.Permutation>,
                       ILocalSearchPathContext<Encodings.PermutationEncoding.Permutation> {

    [StorableConstructor]
    protected ExhaustiveHillClimb(bool deserializing) : base(deserializing) { }
    protected ExhaustiveHillClimb(ExhaustiveHillClimb<TContext> original, Cloner cloner) : base(original, cloner) { }
    public ExhaustiveHillClimb() {
      Name = ItemName;
      Description = ItemDescription;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ExhaustiveHillClimb<TContext>(this, cloner);
    }

    public void Optimize(TContext context) {
      if (double.IsNaN(context.Solution.Fitness))
        context.Evaluate(context.Solution, CancellationToken.None);
      var quality = context.Solution.Fitness;
      try {
        var path = new List<Tuple<Encodings.PermutationEncoding.Permutation, double>>();
        path.Add(Tuple.Create((Encodings.PermutationEncoding.Permutation)context.Solution.Solution.Clone(), quality));

        var result = Exhaustive.HillClimb(context.Random, context.Solution.Solution, quality,
          context.Maximization, context.Evaluate, CancellationToken.None);

        Tuple<Encodings.PermutationEncoding.Permutation, double, int> last = null;
        foreach (var step in result) {
          path.Add(Tuple.Create((Encodings.PermutationEncoding.Permutation)step.Item1.Clone(), step.Item2));
          last = step;
        }
        context.LocalSearchPaths.AddPath(path);
        context.IncrementEvaluatedSolutions(last.Item3);
        context.Iterations = path.Count - 2;
      } finally {
        context.Solution.Fitness = quality;
      }
    }
  }
}
